# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import pandas as pd

try:
    pd.set_option("future.no_silent_downcasting", True)
except pd.errors.OptionError:
    pass



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0
"""
The TabularData class is a wrapper around Dict[str, Union[pd.DataFrame, np.ndarray]],
which supports unit conversions and value substitutions
"""

from typing import Callable, Dict, Generator, Iterable, Optional, Tuple, Union

import numpy as np
import pandas as pd
import structlog

from power_grid_model_io.mappings.unit_mapping import UnitMapping
from power_grid_model_io.mappings.value_mapping import ValueMapping

LazyDataFrame = Callable[[], pd.DataFrame]


class TabularData:
    """
    The TabularData class is a wrapper around Dict[str, Union[pd.DataFrame, np.ndarray]],
    which supports unit conversions and value substitutions
    """

    def __init__(self, **tables: Union[pd.DataFrame, np.ndarray, LazyDataFrame]):
        """
        Tabular data can either be a collection of pandas DataFrames and/or numpy structured arrays.
        The key word arguments will define the keys of the data.

        tabular_data = TabularData(foo=foo_data)
        tabular_data["foo"] --> foo_data

        Args:
            **tables: A collection of pandas DataFrames and/or numpy structured arrays
        """
        for table_name, table_data in tables.items():
            if not isinstance(table_data, (pd.DataFrame, np.ndarray)) and not callable(table_data):
                raise TypeError(
                    f"Invalid data type for table '{table_name}'; "
                    f"expected a pandas DataFrame or NumPy array, got {type(table_data).__name__}."
                )
        self._data: Dict[str, Union[pd.DataFrame, np.ndarray, LazyDataFrame]] = tables
        self._units: Optional[UnitMapping] = None
        self._substitution: Optional[ValueMapping] = None
        self._log = structlog.get_logger(type(self).__name__)

    def set_unit_multipliers(self, units: UnitMapping) -> None:
        """
        Define unit multipliers.

        Args:
            units: A UnitMapping object defining all the units and their conversions (e.g. 1 MW = 1_000_000 W)
        """
        self._units = units

    def set_substitutions(self, substitution: ValueMapping) -> None:
        """
        Define value substitutions

        Args:
            substitution: A ValueMapping defining all value substitutions (e.g. "yes" -> 1)
        """
        self._substitution = substitution

    def get_column(self, table_name: str, column_name: str) -> pd.Series:
        """
        Select a column from a table, while applying unit conversions and value substitutions

        Args:
            table_name: The name of the table as supplied in the constructor
            column_name: The name of the column or "index" to get the index

        Returns:
            The required column, with unit conversions and value substitutions applied
        """
        table_data = self[table_name]

        # If the index 'column' is requested, but no column called 'index' exist,
        # return the index of the dataframe as if it were an actual column.
        if column_name == "index" and "index" not in table_data and hasattr(table_data, "index"):
            return pd.Series(table_data.index, name="index", copy=False)

        column_data = table_data[column_name]

        if isinstance(column_data, np.ndarray):
            column_data = pd.Series(column_data, name=column_name, copy=False)

        # If unit information is available, convert the unit
        if not isinstance(column_data, pd.Series):
            column_data = self._apply_unit_conversion(table_data=table_data, table=table_name, field=column_name)
            if not isinstance(column_data, pd.Series):
                raise TypeError(
                    f"The '{column_name}' column should now be unitless, "
                    f"but it still contains a unit: {column_data.columns.values}"
                )

        return self._apply_value_substitution(column_data=column_data, table=table_name, field=column_name)

    def _apply_value_substitution(self, column_data: pd.Series, table: str, field: str) -> pd.Series:
        if self._substitution is None:  # No substitution defined, at all
            return column_data

        # Find substitutions, ignore if none is found
        try:
            substitutions = self._substitution.get_substitutions(attr=field, table=table)
        except KeyError:
            return column_data

        def sub(value):
            try:
                return substitutions[value]
            except KeyError as ex:
                raise KeyError(
                    f"Unknown substitution for value '{value}' in column '{field}' in table '{table}'"
                ) from ex

        # Apply substitutions
        self._log.debug("Apply value substitutions", table=table, field=field, substitutions=substitutions)
        return column_data.map(sub)

    def _apply_unit_conversion(self, table_data: pd.DataFrame, table: str, field: str) -> pd.Series:
        unit = table_data[field].columns[0]

        try:
            if self._units is None:
                raise KeyError(unit)
            multiplier, si_unit = self._units.get_unit_multiplier(unit)
        except KeyError as ex:
            raise KeyError(f"Unknown unit '{unit}' for column '{field}' in table '{table}'") from ex

        if unit == si_unit:
            self._log.debug("No unit conversion needed", table=table, field=field, unit=unit)
        else:
            self._log.debug(
                "Apply unit conversion", table=table, field=field, unit=unit, si_unit=si_unit, multiplier=multiplier
            )
            try:
                table_data[field] *= multiplier
            except TypeError as ex:
                self._log.error(
                    "Failed to apply unit conversion; the column is not numerical.",
                    table=table,
                    field=field,
                    unit=unit,
                    si_unit=si_unit,
                    multiplier=multiplier,
                    ex=str(ex),
                )
                si_unit = ""

            # Replace the unit with the SI unit
            table_data.columns = table_data.columns.values
            table_data.columns = pd.MultiIndex.from_tuples(table_data.rename(columns={(field, unit): (field, si_unit)}))

        return table_data[pd.MultiIndex.from_tuples([(field, si_unit)])[0]]

    def __len__(self) -> int:
        """
        Return the number of tables (regardless of if they are already loaded or not)

        Returns: The number of tables
        """
        return len(self._data)

    def __contains__(self, table_name: str) -> bool:
        """
        Mimic the dictionary 'in' operator

        Args:
            table_name: The name of the table as supplied in the constructor.

        Returns: True if the table name was supplied in the constructor.
        """
        return table_name in self._data

    def __getitem__(self, table_name: str) -> Union[pd.DataFrame, np.ndarray]:
        """
        Mimic the dictionary [] operator. It returns the 'raw' table data as stored in memory. This can be either a
        pandas DataFrame or a numpy structured array. It is possible that some unit conversions have been applied by
        previous calls to get_column().

        Args:
            table_name: The name of the table as supplied in the constructor

        Returns: The 'raw' table data
        """
        data = self._data[table_name]
        if callable(data):
            self._data[table_name] = data()
        return self._data[table_name]

    def keys(self) -> Iterable[str]:
        """
        Mimic the dictionary .keys() function

        Returns: An iterator over all table names as supplied in the constructor.
        """

        return self._data.keys()

    def items(self) -> Generator[Tuple[str, Union[pd.DataFrame, np.ndarray]], None, None]:
        """
        Mimic the dictionary .items() function

        Returns: A generator of the table names and the raw table data
        """

        # Note: PyCharm complains about the type, but it is correct, as an ItemsView extends from
        # AbstractSet[Tuple[_KT_co, _VT_co]], which actually is compatible with Iterable[_KT_co, _VT_co]
        for key in self._data:
            yield key, self[key]



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0
"""
Value substitution helper class
"""

from typing import Dict, Optional, Union

from power_grid_model_io.mappings.field_mapping import FieldMapping

Value = Union[int, float, str, bool]

#             attr      key    value
Values = Dict[str, Dict[Value, Value]]


# pylint: disable=too-few-public-methods
class ValueMapping(FieldMapping[Dict[Value, Value]]):
    """
    Value substitution helper class
    """

    def __init__(self, mapping: Optional[Values] = None):
        super().__init__(mapping=mapping)
        if mapping is not None:
            self._log.debug(
                "Set value mapping", n_attributes=len(mapping), n_mappings=sum(len(m) for m in mapping.values())
            )

    def get_substitutions(self, attr: str, table: Optional[str] = None) -> Dict[Value, Value]:
        """
        Find the substitutions for a given attribute.
        """
        return self._get_mapping(attr=attr, table=table)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0
"""
Unit mapping helper class
"""

from numbers import Number
from typing import Dict, Optional, Set, Tuple

import numpy as np
import structlog

#            si-unit   unit factor
Units = Dict[str, Optional[Dict[str, float]]]


class UnitMapping:
    """
    Unit mapping helper class.
    The input data is expected to be of the form:
    {
        "A" : None,
        "W": {
            "kW": 1000.0,
            "MW": 1000000.0
        }
     }
    """

    def __init__(self, mapping: Optional[Units] = None):
        self._log = structlog.get_logger(type(self).__name__)
        self._si_units: Set[str] = set()
        self._mapping: Dict[str, Tuple[float, str]] = {}
        self.set_mapping(mapping=mapping if mapping is not None else {})

    def set_mapping(self, mapping: Units):
        """
        Creates an internal mapping lookup table based on input data of the form:
        mapping = {
            "A" : None,
            "W": {
                "kW": 1000.0,
                "MW": 1000000.0
            }
         }
        """
        self._si_units = set(mapping.keys())
        self._mapping = {}
        for si_unit, multipliers in mapping.items():
            if not multipliers:
                continue
            for unit, multiplier in multipliers.items():
                if unit in self._mapping:
                    multiplier_, si_unit_ = self._mapping[unit]
                    raise ValueError(
                        f"Multiple unit definitions for '{unit}': "
                        f"1{unit} = {multiplier_}{si_unit_} = {multiplier}{si_unit}"
                    )
                self._mapping[unit] = (multiplier, si_unit)
                if unit == si_unit:
                    if not np.isclose(multiplier, 1.0, rtol=1.0e-9, atol=0.0):
                        raise ValueError(
                            f"Invalid unit definition for '{unit}': 1{unit} cannot be {multiplier}{si_unit}"
                        )
                    continue  # pragma: no cover (bug in python 3.9)
                self._mapping[unit] = (multiplier, si_unit)
        self._log.debug(
            "Set unit definitions", n_units=len(self._si_units | self._mapping.keys()), n_si_units=len(self._si_units)
        )

    def get_unit_multiplier(self, unit: str) -> Tuple[float, str]:
        """
        Find the correct unit multiplier and the corresponding SI unit
        """
        if unit in self._si_units:
            return 1.0, unit

        multiplier, si_unit = self._mapping[unit]

        if not isinstance(multiplier, Number):
            raise TypeError(f"The multiplier ({multiplier}) for unit '{unit}' is not numerical.")

        return multiplier, si_unit



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import json
from functools import lru_cache
from pathlib import Path
from typing import List, Tuple

import pandas as pd
import pytest
from power_grid_model.data_types import SingleDataset

from power_grid_model_io.converters import VisionExcelConverter
from power_grid_model_io.data_types import ExtraInfo
from power_grid_model_io.utils.json import JsonEncoder

from ..utils import compare_extra_info, component_attributes, component_objects, load_json_single_dataset, select_values

DATA_PATH = Path(__file__).parents[2] / "data" / "vision"
SOURCE_FILE = DATA_PATH / "vision_{language:s}.xlsx"
VALIDATION_FILE = DATA_PATH / "pgm_input_data_{language:s}.json"
LANGUAGES = ["en", "nl"]
VALIDATION_EN = Path(str(VALIDATION_FILE).format(language="en"))
CUSTOM_MAPPING_FILE = DATA_PATH / "vision_9_5_{language:s}.yaml"


@lru_cache
def vision_excel_converter(language: str) -> VisionExcelConverter:
    """
    Read the excel file
    """
    source_file = Path(str(SOURCE_FILE).format(language=language))
    return VisionExcelConverter(source_file, language=language)


@lru_cache
def load_and_convert_excel_file(language: str) -> Tuple[SingleDataset, ExtraInfo]:
    """
    Convert the excel file
    """
    data, extra_info = vision_excel_converter(language=language).load_input_data()
    return data, extra_info


@lru_cache
def load_validation_data(language: str) -> Tuple[SingleDataset, ExtraInfo]:
    """
    Load the validation data from the json file
    """
    validation_file = Path(str(VALIDATION_FILE).format(language=language))
    data, extra_info = load_json_single_dataset(validation_file, data_type="input")
    return data, extra_info


@pytest.fixture
def input_data(request) -> Tuple[SingleDataset, SingleDataset]:
    """
    Read the excel file and the json file, and return the input_data
    """
    actual, _ = load_and_convert_excel_file(language=request.param)
    expected, _ = load_validation_data(language=request.param)
    return actual, expected


@pytest.fixture
def extra_info(request) -> Tuple[ExtraInfo, ExtraInfo]:
    """
    Read the excel file and the json file, and return the extra_info
    """
    _, actual = load_and_convert_excel_file(language=request.param)
    _, expected = load_validation_data(language=request.param)
    return actual, expected


@pytest.mark.parametrize("input_data", LANGUAGES, indirect=True)
def test_input_data(input_data: Tuple[SingleDataset, SingleDataset]):
    """
    Unit test to preload the expected and actual data
    """
    # Arrange
    actual, expected = input_data
    # Assert
    assert len(expected) <= len(actual)


def test_input_data_custom_yaml():
    """
    Unit test to preload the expected and actual data, using a different mapping file other than the one in the default location
    """
    for language in LANGUAGES:
        # Arrange
        actual, _ = VisionExcelConverter(
            Path(str(SOURCE_FILE).format(language=language)),
            language=language,
            mapping_file=Path(str(CUSTOM_MAPPING_FILE).format(language=language)),
        ).load_input_data()
        expected, _ = load_validation_data(language=language)
        # Assert
        assert len(expected) <= len(actual)


@pytest.mark.parametrize(("component", "attribute"), component_attributes(VALIDATION_EN, data_type="input"))
@pytest.mark.parametrize("input_data", LANGUAGES, indirect=True)
def test_attributes(input_data: Tuple[SingleDataset, SingleDataset], component: str, attribute: str):
    """
    For each attribute, check if the actual values are consistent with the expected values
    """
    # Arrange
    actual_data, expected_data = input_data

    # Act
    actual_values, expected_values = select_values(actual_data, expected_data, component, attribute)

    # Assert
    pd.testing.assert_series_equal(actual_values, expected_values)


@pytest.mark.parametrize(
    ("component", "obj_ids"),
    (pytest.param(component, objects, id=component) for component, objects in component_objects(VALIDATION_EN)),
)
@pytest.mark.parametrize("extra_info", LANGUAGES, indirect=True)
def test_extra_info(extra_info: Tuple[ExtraInfo, ExtraInfo], component: str, obj_ids: List[int]):
    """
    For each object, check if the actual extra info is consistent with the expected extra info
    """
    # Arrange
    actual, expected = extra_info

    # Assert
    errors = compare_extra_info(actual=actual, expected=expected, component=component, obj_ids=obj_ids)

    # Raise a value error, containing all the errors at once
    if errors:
        raise ValueError("\n" + "\n".join(errors))


@pytest.mark.parametrize("extra_info", LANGUAGES, indirect=True)
def test_extra_info__serializable(extra_info):
    # Arrange
    actual, _ = extra_info

    # Assert
    json.dumps(actual, cls=JsonEncoder)  # expect no exception


@pytest.mark.parametrize(("language", "table", "column"), [("en", "Nodes", "Number"), ("nl", "Knooppunten", "Nummer")])
def test_get_node_id(language: str, table: str, column: str):
    # Arrange
    converter = vision_excel_converter(language=language)
    _, extra_info = load_and_convert_excel_file(language=language)

    assert converter._source is not None
    source_data = converter._source.load()[table][column]

    # Act/Assert
    for number in source_data:
        pgm_id = converter.get_node_id(number=number)
        assert extra_info[pgm_id]["id_reference"] == {"table": table, "key": {column: number}}


@pytest.mark.parametrize(
    ("language", "table", "column"),
    [
        ("en", "Cables", "Number"),
        ("en", "Lines", "Number"),
        ("en", "Reactance coils", "Number"),
        ("en", "Links", "Number"),
        ("en", "Transformers", "Number"),
        ("en", "Special transformers", "Number"),
        ("en", "Three winding transformers", "Number"),
        ("nl", "Kabels", "Nummer"),
        ("nl", "Verbindingen", "Nummer"),
        ("nl", "Smoorspoelen", "Nummer"),
        ("nl", "Links", "Nummer"),
        ("nl", "Transformatoren", "Nummer"),
        ("nl", "Speciale transformatoren", "Nummer"),
        ("nl", "Driewikkelingstransformatoren", "Nummer"),
    ],
)
def test_get_branch_id(language: str, table: str, column: str):
    # Arrange
    converter = vision_excel_converter(language=language)
    _, extra_info = load_and_convert_excel_file(language=language)

    assert converter._source is not None
    source_data = converter._source.load()[table][column]

    # Act/Assert
    for number in source_data:
        pgm_id = converter.get_branch_id(table=table, number=number)
        assert extra_info[pgm_id]["id_reference"] == {"table": table, "key": {column: number}}


@pytest.mark.parametrize(
    ("language", "table", "columns"),
    [
        ("en", "Loads", ["Node.Number", "Subnumber"]),
        ("en", "Synchronous generators", ["Node.Number", "Subnumber"]),
        ("en", "Wind turbines", ["Node.Number", "Subnumber"]),
        ("en", "Pvs", ["Node.Number", "Subnumber"]),
        ("en", "Sources", ["Node.Number", "Subnumber"]),
        ("en", "Zigzag transformers", ["Node.Number", "Subnumber"]),
        ("en", "Capacitors", ["Node.Number", "Subnumber"]),
        ("en", "Reactors", ["Node.Number", "Subnumber"]),
        ("nl", "Belastingen", ["Knooppunt.Nummer", "Subnummer"]),
        ("nl", "Synchrone generatoren", ["Knooppunt.Nummer", "Subnummer"]),
        ("nl", "Windturbines", ["Knooppunt.Nummer", "Subnummer"]),
        ("nl", "Pv's", ["Knooppunt.Nummer", "Subnummer"]),
        ("nl", "Netvoedingen", ["Knooppunt.Nummer", "Subnummer"]),
        ("nl", "Nulpuntstransformatoren", ["Knooppunt.Nummer", "Subnummer"]),
        ("nl", "Condensatoren", ["Knooppunt.Nummer", "Subnummer"]),
        ("nl", "Spoelen", ["Knooppunt.Nummer", "Subnummer"]),
    ],
)
def test_get_get_appliance_id(language: str, table: str, columns: List[str]):
    # Arrange
    converter = vision_excel_converter(language=language)
    _, extra_info = load_and_convert_excel_file(language=language)

    assert converter._source is not None
    source_data = converter._source.load()[table][columns]

    # Act/Assert
    assert isinstance(source_data, pd.DataFrame)
    for _, (node_number, sub_number) in source_data.iterrows():
        pgm_id = converter.get_appliance_id(table=table, node_number=node_number, sub_number=sub_number)
        assert extra_info[pgm_id]["id_reference"] == {
            "table": table,
            "key": {columns[0]: node_number, columns[1]: sub_number},
        }


@pytest.mark.parametrize(
    ("language", "table", "name", "columns"),
    [
        ("en", "Transformer loads", "transformer", ["Node.Number", "Subnumber"]),
        ("en", "Transformer loads", "internal_node", ["Node.Number", "Subnumber"]),
        ("en", "Transformer loads", "load", ["Node.Number", "Subnumber"]),
        ("en", "Transformer loads", "generation", ["Node.Number", "Subnumber"]),
        ("en", "Transformer loads", "pv_generation", ["Node.Number", "Subnumber"]),
        ("nl", "Transformatorbelastingen", "transformer", ["Knooppunt.Nummer", "Subnummer"]),
        ("nl", "Transformatorbelastingen", "internal_node", ["Knooppunt.Nummer", "Subnummer"]),
        ("nl", "Transformatorbelastingen", "load", ["Knooppunt.Nummer", "Subnummer"]),
        ("nl", "Transformatorbelastingen", "generation", ["Knooppunt.Nummer", "Subnummer"]),
        ("nl", "Transformatorbelastingen", "pv_generation", ["Knooppunt.Nummer", "Subnummer"]),
    ],
)
def test_get_get_virtual_id(language: str, table: str, name: str, columns: List[str]):
    # Arrange
    converter = vision_excel_converter(language=language)
    _, extra_info = load_and_convert_excel_file(language=language)

    assert converter._source is not None
    source_data = converter._source.load()[table][columns]

    # Act/Assert
    assert isinstance(source_data, pd.DataFrame)
    for _, (node_number, sub_number) in source_data.iterrows():
        pgm_id = converter.get_virtual_id(table=table, obj_name=name, node_number=node_number, sub_number=sub_number)
        assert extra_info[pgm_id]["id_reference"] == {
            "table": table,
            "name": name,
            "key": {columns[0]: node_number, columns[1]: sub_number},
        }



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import pytest
from pytest import fixture

from power_grid_model_io.mappings.multiplier_mapping import MultiplierMapping


@fixture
def mapping() -> MultiplierMapping:
    return MultiplierMapping(
        {
            "sources.p": 1.0,
            ".*_power": 2.0,
            "Q": 3.0,
        }
    )


def test_get_multiplier__exact_match(mapping: MultiplierMapping):
    assert mapping.get_multiplier("Q") == 3.0


def test_get_multiplier__exact_match_with_table(mapping: MultiplierMapping):
    assert mapping.get_multiplier("p", table="sources") == 1.0
    with pytest.raises(KeyError):
        mapping.get_multiplier("p")


def test_get_multiplier__regex_match(mapping: MultiplierMapping):
    assert mapping.get_multiplier("measured_power") == 2.0


def test_get_multiplier__no_match(mapping: MultiplierMapping):
    with pytest.raises(KeyError):
        mapping.get_multiplier("X")



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0
"""
The json file store
"""

import json
from pathlib import Path
from typing import Optional

from power_grid_model_io.data_stores.base_data_store import BaseDataStore
from power_grid_model_io.data_types import StructuredData
from power_grid_model_io.utils.json import JsonEncoder, compact_json_dump


class JsonFileStore(BaseDataStore[StructuredData]):
    """
    The json file store expects each json file to be eiter a single dictionary, or a list of dictonaries.
    """

    __slots__ = ("_indent", "_compact", "_file_path")

    def __init__(self, file_path: Path):
        super().__init__()
        self._indent: Optional[int] = 2
        self._compact: bool = True
        self._file_path: Path = Path(file_path)

        # Check JSON file name
        if file_path.suffix.lower() != ".json":
            raise ValueError(f"JsonFile file should be a .json file, {file_path.suffix} provided.")

    def set_indent(self, indent: Optional[int]) -> None:
        """
        Change the number of spaces used for each indent level (affects output only)

        Args:
            indent: Number of spaces for each indent level. None = all json on a single line.
        """
        self._indent = indent

    def set_compact(self, compact: bool) -> None:
        """
        In compact mode, each object will be output on a single line. Note that the JsonFileStore is not very
        general; it assumes that data is either a dictionary of this format:

        {
          "category_0":
            [
              {"attribute_0": ..., "attribute_1": ..., ...},
              {"attribute_0": ..., "attribute_1": ..., ...},
            ],
          "category_1":
            [
              {"attribute_0": ..., "attribute_1": ..., ...},
              {"attribute_0": ..., "attribute_1": ..., ...},
            ],
          ...
        }

        or a list of those dictionaries.

        Args:
            compact: Boolean defining if the output should be stored compact or not
        """
        self._compact = compact

    def load(self) -> StructuredData:
        """
        Loads a JSON file, validates the structure and returns it in a native python data format

        Returns: StructuredData
        """
        with self._file_path.open(mode="r", encoding="utf-8") as file_pointer:
            data = json.load(file_pointer)
        self._validate(data=data)
        return data

    def save(self, data: StructuredData) -> None:
        """
        Saves the native python data format as a JSON file

        Args:
            data: StructuredData
        """
        self._validate(data=data)
        self._file_path.parent.mkdir(parents=True, exist_ok=True)
        with self._file_path.open(mode="w", encoding="utf-8") as file_pointer:
            if self._compact and self._indent:
                max_level = 3
                if isinstance(data, list):
                    max_level += 1
                compact_json_dump(data, file_pointer, indent=self._indent, max_level=max_level)
            else:
                json.dump(data, file_pointer, indent=self._indent, cls=JsonEncoder)

    def _validate(self, data: StructuredData) -> None:
        # The data should be either a dictionary, or a (possibly empty) list of dictionaries
        if not isinstance(data, (dict, list)):
            raise TypeError(f"Invalid data type for {type(self).__name__}: {type(data).__name__}")

        if isinstance(data, list) and any(not isinstance(x, dict) for x in data):
            type_names = sorted({type(x).__name__ for x in data})
            if len(type_names) == 1:
                type_str = type_names.pop()
            else:
                type_str = "Union[" + ", ".join(type_names) + "]"
            raise TypeError(f"Invalid data type for {type(self).__name__}: List[{type_str}]")



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0
"""
Abstract data store class
"""

from abc import ABC, abstractmethod
from typing import Generic, TypeVar

import structlog

T = TypeVar("T")


class BaseDataStore(Generic[T], ABC):
    """
    Abstract data store class
    """

    def __init__(self):
        """
        Initialize a logger
        """
        self._log = structlog.get_logger(type(self).__name__)

    @abstractmethod  # pragma: no cover
    def load(self) -> T:
        """
        The method that loads the data from one or more sources and returns it in the specified format.
        Note that the load() method does not receive a reference to the data source(s); i.e. the data source(s)
        should be set in the constructor, or in a separate member method.

        Returns: Loaded data of type <T>
        """

    @abstractmethod  # pragma: no cover
    def save(self, data: T) -> None:
        """
        The method that saves the data to one or more destinations.
        Note that the save() method does not receive a reference to the data destination(s); i.e. the data
        destination(s) should be set in the constructor, or in a separate member method.

        Args:
            data: Tha data to store shoul dbe of type <T>
        """



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0
"""
Tabular data mapping helper class
"""
from typing import Dict, Generator, List, Tuple, Union

import structlog

AttributeValue = Union[int, float, str, Dict, List]
InstanceAttributes = Dict[str, AttributeValue]
Components = Dict[str, Union[InstanceAttributes, List[InstanceAttributes]]]
Tables = Dict[str, Components]


class TabularMapping:
    """
    Tabular data mapping helper class
    """

    def __init__(self, mapping: Tables):
        self._log = structlog.get_logger(type(self).__name__)
        self._log.debug("Set tabular mapping", n_mappings=len(mapping))
        self._mapping: Tables = mapping

    def tables(self) -> Generator[str, None, None]:
        """
        Return the names of the tables (as a generator)

        Yields:
            table_name
        """
        return (key for key in self._mapping.keys())

    def instances(self, table: str) -> Generator[Tuple[str, InstanceAttributes], None, None]:
        """
        Return instance definitions (as a generator)

        Yields:
            component_name, instance_attribute_mapping
        """
        table_mapping = self._mapping.get(table, {})
        if not isinstance(table_mapping, dict):
            raise TypeError(
                f"Invalid table mapping for {table}; expected a dictionary got {type(table_mapping).__name__}"
            )
        for component, instances in table_mapping.items():
            if not isinstance(instances, list):
                yield component, instances
            else:
                for instance in instances:
                    yield component, instance
